// A map data structure that holds object keys weakly, yet can also hold
// non-object keys, unlike the native WeakMap.
var UniversalWeakMap = function UniversalWeakMap() {
  // Since a WeakMap cannot hold primitive values as keys, we need a
  // backup Map instance to hold primitive keys. Both this._weakMap and
  // this._strongMap are lazily initialized.
  this._weakMap = null;
  this._strongMap = null;
};

// Since get and set are the only methods used, that's all I've
// implemented here.

UniversalWeakMap.prototype.get = function get (key) {
  var map = this._getMap(key, false);
  if (map) {
    return map.get(key);
  }
};

UniversalWeakMap.prototype.set = function set (key, value) {
  this._getMap(key, true).set(key, value);
  // An actual Map or WeakMap would return this here, but returning the
  // value is more convenient below.
  return value;
};

UniversalWeakMap.prototype._getMap = function _getMap (key, canCreate) {
  if (! canCreate) {
    return isObjRef(key) ? this._weakMap : this._strongMap;
  }
  if (isObjRef(key)) {
    return this._weakMap || (this._weakMap = new WeakMap);
  }
  return this._strongMap || (this._strongMap = new Map);
};

function isObjRef(value) {
  switch (typeof value) {
  case "object":
    if (value === null) {
      return false;
    }
  case "function":
    return true;
  default:
    return false;
  }
}

var useSymbol = typeof Symbol === "function";

var brand = useSymbol
  ? Symbol.for("immutable-tuple")
  : "@@__IMMUTABLE_TUPLE__@@";

var globalKey = useSymbol
  ? Symbol.for("immutable-tuple-root")
  : "@@__IMMUTABLE_TUPLE_ROOT__@@";

// The mustConvertThisToArray value is true if the corresponding Array
// method will not attempt to modify `this`, which means we can pass a
// Tuple object as `this` without first converting it to an Array.
function forEachArrayMethod(fn) {
  function call(name, mustConvertThisToArray) {
    var desc = Object.getOwnPropertyDescriptor(Array.prototype, name);
    fn(name, desc, !! mustConvertThisToArray);
  }
  call("slice");
  call("includes");
  call("indexOf");
  call("forEach");
  call("filter");
  call("map");
  call("every");
  call("some");
  call("reduce");
  call("reduceRight");
  call("toString");
  call("toLocaleString");
  call("join");
  call("reverse", true);
  call("sort", true);
  call("lastIndexOf");
  call("find");
  call("findIndex");
  call(useSymbol && Symbol.iterator || "@@iterator");
}

// If this package is installed multiple times, there could be mutiple
// implementations of the tuple function with distinct tuple.prototype
// objects, but the shared pool of tuple objects must be the same across
// all implementations.
var root = globalKey in global
  ? global[globalKey]
  : def(global, globalKey, new UniversalWeakMap, false);

var ref = Array.prototype;
var concat = ref.concat;
var reusableTempArray = [];

function tuple() {
  var items = [], len = arguments.length;
  while ( len-- ) items[ len ] = arguments[ len ];

  return intern(items);
}

function intern(array) {
  var node = root;

  array.forEach(function (item) {
    node = node.get(item) || node.set(item, new UniversalWeakMap);
  });

  if (node.tuple) {
    return node.tuple;
  }

  var t = Object.create(tuple.prototype);

  array.forEach(function (item, i) { return def(t, i, item, true); });
  def(t, "length", array.length, false);

  return node.tuple = t;
}

function def(obj, name, value, enumerable) {
  Object.defineProperty(obj, name, {
    value: value,
    enumerable: !! enumerable,
    writable: false,
    configurable: false
  });
  return value;
}

function isTuple(that) {
  return !! (that && that[brand] === true);
}

tuple.isTuple = isTuple;

// Turn this Tuple into an ordinary array, optionally reusing an
// existing array.
tuple.prototype.toArray = function (array) {
  var this$1 = this;
  if ( array === void 0 ) array = [];

  var ref = this;
  var length = ref.length;
  for (var i = 0; i < length; ++i) {
    var item = this$1[i];
    array[i] = isTuple(item) ? item.toArray() : item;
  }
  array.length = length;
  return array;
};

// Like Array.prototype.concat, except that extra effort is required to
// convert any tuple arguments to arrays, so that Array.prototype.concat
// will do the right thing.
tuple.prototype.concat = function () {
  var args = [], len = arguments.length;
  while ( len-- ) args[ len ] = arguments[ len ];

  return intern(concat.apply(
    this.toArray(reusableTempArray),
    args.map(function (item) { return isTuple(item) ? item.toArray() : item; })
  ));
};

def(tuple.prototype, brand, true, false);

forEachArrayMethod(function (name, desc, mustConvertThisToArray) {
  var method = desc && desc.value;
  if (typeof method === "function") {
    desc.value = function () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      var result = method.apply(
        mustConvertThisToArray
          ? this.toArray(reusableTempArray)
          : this,
        args
      );
      return Array.isArray(result) ? intern(result) : result;
    };
    Object.defineProperty(tuple.prototype, name, desc);
  }
});

export default tuple;
export { tuple };
